@@ -1,6 +1,6 @@
 /*
 
-Copyright (C) 2002 - 2004 Christian Kreibich <christian@whoop.org>.
+Copyright (C) 2002 - 2006 Christian Kreibich <christian@whoop.org>.
 
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to
@@ -22,6 +22,10 @@ IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
 */
+#ifdef HAVE_CONFIG_H
+#  include <config.h>
+#endif
+
 #include <libnd_globals.h>
 
 char                  libnd_pcap_errbuf[PCAP_ERRBUF_SIZE];
