@@ -1,6 +1,6 @@
 /*
 
-Copyright (C) 2000 - 2004 Christian Kreibich <christian@whoop.org>.
+Copyright (C) 2000 - 2006 Christian Kreibich <christian@whoop.org>.
 
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to
@@ -31,6 +31,7 @@ CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 #include <stdlib.h>
 #include <stdio.h>
 #include <sys/types.h>
+#include <errno.h>
 #ifdef LINUX
 #define __FAVOR_BSD
 #endif
@@ -258,10 +259,19 @@ libnd_packet_dump(const LND_Packet *packet, pcap_dumper_t *dumper)
   if (!packet || !dumper)
     return;
   
+  /*
+   * if (fwrite(&packet->ph, sizeof(packet->ph), 1, f) != 1)
+   * fprintf(stderr, "WARNING: packet dump failed, %s\n", strerror(errno));
+   *
+   * if (fwrite(packet->data, packet->ph.caplen, 1, f) != 1)
+   *    fprintf(stderr, "WARNING: packet dump failed, %s\n", strerror(errno));
+   */
+  
   /* Dammit, pcap doesn't return the outcome of the
    * operation! We could run out of diskspace here and
    * never know ... big fat FIXME for the future.
    */
+  
   pcap_dump((u_char*) dumper, &packet->ph, packet->data);
 }
 
@@ -292,7 +302,8 @@ libnd_packet_duplicate(LND_Packet *p)
   if (!p)
     return (NULL);
 
-  copy = libnd_packet_new(p->part, p->ph.caplen);
+  /* copy = libnd_packet_new(p->part, p->ph.caplen);*/
+  copy = libnd_packet_new(NULL, p->ph.caplen);
 
   copy->ph          = p->ph;
   copy->protocols   = p->protocols;
@@ -329,6 +340,7 @@ packet_init(LND_Packet *packet, pcap_t * pcap)
   int type;
   LND_Protocol *proto;
   LND_Trace *trace;
+  guchar *data_end, *used_data_end;
 
   D_ENTER;
 
@@ -362,20 +374,25 @@ packet_init(LND_Packet *packet, pcap_t * pcap)
 
   /* Start initializing the packet using the protocol we've just
    * determined. Each protocol decides what protocol to use next,
-   * if applicable.
+   * if applicable. Should not all data end up being used, we
+   * present those in raw hex.
    */
-  proto->init_packet(packet, packet->data, libnd_packet_get_end(packet));
-
+  data_end = libnd_packet_get_end(packet);
+  
+  if ( (used_data_end =
+	proto->init_packet(packet, packet->data, data_end)) < data_end)
+    libnd_raw_proto_get()->init_packet(packet, used_data_end, data_end);
+  
   /* Depending on our current filter configuration, mark this packet
    * as filtered or not. Only applicable if the packet belongs to trace,
    * as the trace defines the filtering configuration.
    */
   if ( (trace = libnd_packet_get_trace(packet)))
     libnd_filter_list_apply(trace->filters, packet, trace->filter_mode);
-
+  
   /* Finally inform anyone interested that we've initialized this packet. */
   libnd_packet_tell_observers(packet, LND_PACKET_INITIALIZED, NULL);
-
+  
   D_RETURN;
 }
 
@@ -412,7 +429,7 @@ libnd_packet_cleanup(LND_Packet *packet)
   /* Remove any existing offsets */
   if (packet->pd)
     {
-      g_list_foreach(packet->pd, packet_free_proto_data, libnd_packet_get_trace(packet));
+      g_list_foreach(packet->pd, packet_free_proto_data, NULL);
       g_list_free(packet->pd);
       packet->pd = NULL;
     }
@@ -523,7 +540,7 @@ libnd_packet_update(LND_Packet *packet, LND_Protocol *proto, guint nesting)
   /* Make sure the current data doesn't get
      deleted, and clean up the rest. */
   l->data = NULL;
-  g_list_foreach(l, packet_free_proto_data, libnd_packet_get_trace(packet));
+  g_list_foreach(l, packet_free_proto_data, NULL);
   g_list_free(l);
 
   D(("Updating packet from %s/%i onward\n",
@@ -540,6 +557,12 @@ libnd_packet_update(LND_Packet *packet, LND_Protocol *proto, guint nesting)
   D_RETURN;
 }
 
+guint
+libnd_packet_get_num_protos(LND_Packet *packet)
+{
+  return g_list_length(packet->pd);
+}
+
 
 guchar *
 libnd_packet_get_data(const LND_Packet *packet,
@@ -549,7 +572,13 @@ libnd_packet_get_data(const LND_Packet *packet,
   LND_ProtoData *pd;
   GList *l;
 
-  if (!packet || !proto || !libnd_packet_has_proto(packet, proto))
+  if (!packet)
+    return NULL;
+
+  if (!proto)
+    return packet->data;
+  
+  if (!libnd_packet_has_proto(packet, proto))
     return NULL;
 
   for (l = packet->pd; l; l = g_list_next(l))
@@ -806,7 +835,8 @@ libnd_packet_foreach_proto(LND_Packet *packet,
     {
       pd = (LND_ProtoData *) l->data;
 
-      callback(packet, pd, user_data);
+      if (! callback(packet, pd, user_data))
+	break;
     }
 }
 
@@ -826,7 +856,8 @@ libnd_packet_foreach_proto_backward(LND_Packet *packet,
     {
       pd = (LND_ProtoData *) l->data;
 
-      callback(packet, pd, user_data);
+      if (! callback(packet, pd, user_data))
+	break;
     }
 }
 
@@ -921,25 +952,39 @@ libnd_packet_get_proto_nesting(const LND_Packet *packet,
 
 typedef struct nd_fix_data
 {
-  gboolean     changed;
+  gboolean      changed;
+  guint         stop_proto; /* Protocol at which to stop fixing. */
+  guint         total_proto; /* Total # protocols in packet */
+  int           delta;  /* Only used for size change operation */
 } LND_FixData;
 
 
-static void
+static gboolean
 packet_fix_cb(LND_Packet *p, LND_ProtoData *pd, void *user_data)
 {
   LND_FixData *data = (LND_FixData *) user_data;
 
-  if (!p || !pd)
-    return;
+  if (!p || !pd) 
+    return TRUE;
 
+  /* If this protocol header is beyond where we want to
+   * correct, skip -- we're iterating backward.
+   */
+  if (--data->total_proto > data->stop_proto)
+    {
+      D(("Skipping fix of protocol header #%i\n", data->total_proto + 1));
+      return TRUE;
+    }
+  
   /* Fix up the protocol data, depending on how the
    * protocol plugin implements that operation.
    * If the packet was modified, set a flag to let our
-   * called know about it.
+   * caller know about it.
    */
   if (pd->inst.proto->fix_packet(p))
     data->changed = TRUE;
+
+  return TRUE;
 }
 
 
@@ -952,6 +997,8 @@ libnd_packet_fix(LND_Packet *packet)
     return FALSE;
 
   memset(&data, 0, sizeof(LND_FixData));
+  data.total_proto = libnd_packet_get_num_protos(packet);
+  data.stop_proto = libnd_packet_get_last_fixable_proto(packet);
 
   /* Iterate over the protocols contained in the packet,
    * from inner- to outermost, and call our callback for
@@ -961,12 +1008,182 @@ libnd_packet_fix(LND_Packet *packet)
 				      packet_fix_cb,
 				      &data);
   
-
   libnd_packet_tell_observers(packet, LND_PACKET_FIXED, NULL);
 
   return data.changed;
 }
 
+static gboolean
+packet_proceed_cb(LND_Packet *p, LND_ProtoData *pd, void *user_data)
+{
+  int *index = (int *) user_data;
+  
+  if (! pd->inst.proto->can_fix_proceed(p, pd))
+    return FALSE;
+  
+  (*index)++;
+  return TRUE;
+}
+
+int
+libnd_packet_get_last_fixable_proto(const LND_Packet *packet)
+{
+  int index = 0;
+  
+  /* First find out whether there's a point in the packet
+   * data beyond which it makes no sense to correct checksums.
+   */
+  libnd_packet_foreach_proto((LND_Packet *) packet, packet_proceed_cb, &index);
+
+  return index;
+}
+
+int
+libnd_packet_get_proto_index(const LND_Packet *packet,
+			     const LND_ProtoInst *pi)
+{
+  GList *l;
+  int index = 0;
+
+  if (!packet || !pi)
+    return -1;
+
+  for (l = packet->pd; l; l = g_list_next(l), index++)
+    {
+      LND_ProtoData *pd = (LND_ProtoData *) l->data;
+
+      if (pd->inst.proto == pi->proto &&
+	  pd->inst.nesting == pi->nesting)
+	return index;
+    }
+
+  return -1;
+}
+
+LND_ProtoData *
+libnd_packet_get_nth_proto(LND_Packet *packet, int n)
+{
+  GList *l;
+
+  if (! packet || n < 0 || libnd_packet_get_num_protos(packet) <= (unsigned) n)
+    return NULL;
+  
+  for (l = packet->pd; l; l = g_list_next(l), n--)
+    {
+      if (n == 0)
+	return (LND_ProtoData*) l->data;
+    }
+  
+  return NULL;  
+}
+
+static gboolean
+packet_adjust_len_cb(LND_Packet *p, LND_ProtoData *pd, void *user_data)
+{
+  LND_FixData *data = (LND_FixData *) user_data;
+
+  if (!p || !pd) 
+    return TRUE;
+
+  if (pd->inst.proto->adjust_len(p, data->delta))
+    data->changed = TRUE;
+
+  return TRUE;
+}
+
+gboolean
+libnd_packet_adjust_len(LND_Packet *packet, int delta)
+{
+  LND_FixData data;
+
+  if (!packet || delta == 0)
+    return TRUE;
+
+  /* We can at most shrink the len to an empty packet. */
+  if (delta < 0)
+    delta = MAX(- (int)packet->ph.len, delta);
+
+  packet->ph.len += delta;
+
+  /* If the packet has shrunk so that the caplen is now
+   * more than the actual packet, we need to update the
+   * caplen as well.
+   */
+  if (packet->ph.caplen > packet->ph.len)
+    /* Note that packet->ph.len - packet->ph.caplen will be negative */
+    libnd_packet_adjust_caplen(packet, packet->ph.len - packet->ph.caplen);
+  
+  /* Give the protocols a chance to update their headers. */
+  data.changed = FALSE;
+  data.delta = delta;
+  libnd_packet_foreach_proto(packet,
+			     packet_adjust_len_cb,
+			     &data);
+  
+  /* To be sure, re-initialize the protocols' pointers into the packet.
+   * (New space should be claimed by a protocol.)
+   */
+  libnd_packet_init(packet);
+  libnd_packet_fix(packet);
+  
+  /* Notify observers of the fact that the size changed. */
+  libnd_packet_tell_observers(packet, LND_PACKET_LEN_CHANGED, &delta);
+  
+  return TRUE;
+}
+
+gboolean
+libnd_packet_adjust_caplen(LND_Packet *packet, int delta)
+{
+  LND_FixData data;
+  LND_Trace *trace;
+
+  if (!packet || delta == 0)
+    return TRUE;
+
+  data.changed = FALSE;
+  data.delta = delta;
+
+  /* If the delta is positive and the increase is more than the wire
+   * packet size, we need to up the latter too.
+   */
+  if (packet->ph.caplen + delta > packet->ph.len)
+    libnd_packet_adjust_len(packet, packet->ph.caplen + delta - packet->ph.len);
+  
+  /* Check whether we have to reallocate packet space, and do so
+   * if necessary.
+   */
+  if (packet->ph.caplen + delta > packet->data_size)
+    {
+      guint newsize = libnd_prec_calc_mem(packet->ph.caplen + delta);
+      guchar *newdata;
+      
+      if (! (newdata = g_renew(guchar, packet->data, newsize)))
+	return FALSE;
+
+      packet->data = newdata;
+      packet->data_size = newsize;
+    }
+
+  /* Initialize the new memory region, if we're growing the caplen. */
+  if (delta > 0)
+    memset(packet->data + packet->ph.caplen, 0, delta);
+  
+  packet->ph.caplen += delta;
+
+  /* If we've increased size to more than the trace's max snaplen,
+   * we need to update the pcap file header as well.
+   */
+  if ( (trace = libnd_packet_get_trace(packet)) &&
+       packet->ph.caplen >= trace->tcpdump.pfh.snaplen)
+    trace->tcpdump.pfh.snaplen = packet->ph.caplen;
+
+  /* To be sure, re-initialize the protocols' pointers into the packet. */
+  libnd_packet_init(packet);
+  libnd_packet_fix(packet);
+  
+  return TRUE;
+}
 
 LND_PacketObserver *
 libnd_packet_observer_new(void)
@@ -1075,7 +1292,17 @@ libnd_packet_tell_observers(LND_Packet *packet, LND_PacketObserverOp op, void *d
 	  if (ob->packet_fixed)
 	    ob->packet_fixed(packet);
 	  break;
-
+	  
+	case LND_PACKET_LEN_CHANGED:
+	  if (ob->packet_len_changed)
+	    ob->packet_len_changed(packet);
+	  break;
+	  
+	case LND_PACKET_CAPLEN_CHANGED:
+	  if (ob->packet_caplen_changed)
+	    ob->packet_caplen_changed(packet);
+	  break;
+	  
 	default:
 	  D(("Unknown packet operation\n"));
 	}
