@@ -1,6 +1,6 @@
 /*
 
-Copyright (C) 2000 - 2004 Christian Kreibich <christian@whoop.org>.
+Copyright (C) 2000 - 2006 Christian Kreibich <christian@whoop.org>.
 
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to
@@ -306,7 +306,7 @@ prefs_domain_add(GList *domains, LND_PrefsDomain *domain)
       return domains;
     }
 
-  D(("Adding domain %s to list, now %i domains\n",
+  D(("Adding new domain %s to list, now %i domains\n",
      domain->name, g_list_length(domains) + 1));
 
   return g_list_prepend(domains, domain);
@@ -321,6 +321,7 @@ prefs_domain_free(LND_PrefsDomain *domain)
   g_free(domain->name);
   prefs_table_free(domain->table);
   g_list_free(domain->apply_callbacks);
+  g_list_free(domain->update_callbacks);
   g_free(domain);
 }
 
@@ -363,8 +364,8 @@ prefs_read_config_file(GList *domains)
   LND_PrefsDomain *domain;
   int     int_data;
   float   flt_data;
-  char    str_data[MAXPATHLEN];
-  char    dom_name[MAXPATHLEN];
+  char    str_data[4096];
+  char    dom_name[512];
   char   *key;
   const char *cfg_file = NULL;
   FILE   *f;
@@ -378,9 +379,12 @@ prefs_read_config_file(GList *domains)
 
   for ( ; ; )
     {
-      if (fscanf(f, "%s %u", dom_name, (int*) &type) == EOF)
+      if (fscanf(f, "%512s %u", dom_name, (int*) &type) == EOF)
 	break;
       
+      /* Terminate string to be extra sure. */
+      dom_name[511] = '\0';
+
       if ( (key = strchr(dom_name, '/')) == NULL)
 	{
 	  D(("Domain/key error in key %s\n", dom_name));
@@ -394,7 +398,7 @@ prefs_read_config_file(GList *domains)
       
       if (!domain)
 	{
-	  D(("Domain %s not found when reading file %s\n",
+	  D(("Domain %s not found when reading %s, allocating.\n",
 	     dom_name, cfg_file));
 	  domain = prefs_domain_new(domains, dom_name, NULL, 0);
 	  domains = prefs_domain_add(domains, domain);
@@ -403,17 +407,34 @@ prefs_read_config_file(GList *domains)
       switch (type)
 	{
 	case LND_PREFS_INT:
-	  fscanf(f, "%i\n", &int_data);
+	  if (fscanf(f, "%i\n", &int_data) != 1)
+	    {
+	      D(("Domain/key error in key %s/%s\n", dom_name, key));
+	      continue;
+	    }
+
 	  prefs_set_item_direct(domain->table, key, LND_PREFS_INT, &int_data);
 	  break;
 
 	case LND_PREFS_FLT:
-	  fscanf(f, "%f\n", &flt_data);
+	  if (fscanf(f, "%f\n", &flt_data) != 1)
+	    {
+	      D(("Domain/key error in key %s/%s\n", dom_name, key));
+	      continue;
+	    }
+
 	  prefs_set_item_direct(domain->table, key, LND_PREFS_FLT, &flt_data);
 	  break;
 
 	case LND_PREFS_STR:
-	  fscanf(f, "%s\n", str_data);
+	  if (fscanf(f, "%s4096\n", str_data) != 1)
+	    {
+	      D(("Domain/key error in key %s/%s\n", dom_name, key));
+	      continue;
+	    }
+
+	  str_data[4095] = '\0';
+
 	  /* Store only if it's not our "undefined" marker */
 	  if (strcmp(str_data, "---"))
 	    prefs_set_item_direct(domain->table, key, LND_PREFS_STR, str_data);
@@ -422,6 +443,9 @@ prefs_read_config_file(GList *domains)
 	default:
 	  D(("Unknown preference entry type for item '%s'.\n", key));
 	}
+      
+      D(("Added key %s to domain %s, size now %i\n",
+	 key, domain->name, g_hash_table_size(domain->table)));
     }
   
   fclose(f);
@@ -556,6 +580,8 @@ libnd_prefs_load(void)
 {
   LND_PrefsDomain *domain;
   
+  D_ENTER;
+
   domain = prefs_domain_new(global_domains, LND_DOM_NETDUDE,
 			    prefs_entries_netdude,
 			    sizeof (prefs_entries_netdude) / sizeof(LND_PrefsEntry));
@@ -572,13 +598,18 @@ libnd_prefs_load(void)
   prefs_write_config_file(global_domains);
 
   libnd_prefs_apply();
+  D_RETURN;
 }
 
 
 int
 libnd_prefs_save(void)
 {
-  return prefs_write_config_file(global_domains);
+  int result;
+
+  D_ENTER;
+  result = prefs_write_config_file(global_domains);
+  D_RETURN_(result);
 }
 
 
@@ -587,9 +618,8 @@ libnd_prefs_apply(void)
 {
   GList *l, *l2;
   
-  libnd_prefs_get_int_item(LND_DOM_NETDUDE, "tcpdump_print_timestamp",
-			   &libnet_tcpdump_print_timestamp);
-  
+  /* First make a pass over all domains, applying the new settings.
+   */
   for (l = global_domains; l; l = g_list_next(l))
     {
       LND_PrefsDomain *domain = (LND_PrefsDomain *) l->data;
@@ -602,6 +632,27 @@ libnd_prefs_apply(void)
 	    callback(domain, NULL);
 	}
     }
+  
+  libnd_prefs_get_int_item(LND_DOM_NETDUDE, "tcpdump_print_timestamp",
+			   &libnd_tcpdump_print_timestamp);
+
+  /* In a second pass, update the application as necessary.
+   * In cases where updates depend on settings from other
+   * domains, we can only now be sure that those have been
+   * applied.
+   */
+  for (l = global_domains; l; l = g_list_next(l))
+    {
+      LND_PrefsDomain *domain = (LND_PrefsDomain *) l->data;
+      
+      for (l2 = domain->update_callbacks; l2; l2 = g_list_next(l2))
+	{
+	  LND_PrefsCallback callback = (LND_PrefsCallback) l2->data;
+	  
+	  if (callback)
+	    callback(domain, NULL);
+	}
+    }  
 }
 
 
@@ -649,6 +700,17 @@ libnd_prefs_domain_add_apply_cb(LND_PrefsDomain *domain,
 }
 
 
+void             
+libnd_prefs_domain_add_update_cb(LND_PrefsDomain *domain,
+				 LND_PrefsCallback update_cb)
+{
+  if (!domain)
+    return;
+  
+  domain->update_callbacks = g_list_append(domain->update_callbacks, update_cb);
+}
+
+
 void           
 libnd_prefs_foreach_domain(LND_PrefsCallback callback, void *user_data)
 {
